SystemOrganization addCategory: #Squeak2JS!

TestCase subclass: #S2JParserTest
	instanceVariableNames: 'currentRule'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 11:58'!
assert: aString matches: aRule
	"In most cases, throws an error when failing (hard fail)"
	
	| result |
	result := S2JParser matchString: aString with: aRule.
	self assert: result notNil.
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:16'!
assert: aString matches: aRule andProduces: anObject
	"In most cases, throws an error when failing (hard fail). #= Checks collections deeply."
	
	| result |
	result := self assert: aString matches: aRule.
	self assert: result = anObject.
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 19:24'!
assert: aString matches: aRule andProducesNode: aCollection
	"Matches the string to the rule and compares the result to an S2JNode containing the given collection and named like the rule.
	(Follows a convention)."
	
	| result |
	result := self assert: aString matches: aRule.
	self assert: result = (S2JNode named: aRule asString with: aCollection).
	^ result! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:18'!
assertFails: aString
	"Checks against the currentRule"
	
	self
		assertFails: aString
		toMatch: self currentRule.! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:23'!
assertFails: aString toMatch: aRule
	
	self
		should: [ self assert: aString matches: aRule ]
		raise: OM2Fail
		description: 'Matching should have failed but didn''t!!'.! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 14:15'!
assertMatches: aString
	"Check against the currentRule"
	
	self assert: aString matches: self currentRule.! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 20:06'!
assertMatches: aString andProduces: anObject
	"Check against the currentRule"
	
	^ self
		assert: aString
		matches: self currentRule
		andProduces: anObject! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 20:06'!
assertMatches: aString andProducesNode: aCollection
	"Check against the currentRule"
	
	^ self
		assert: aString
		matches: self currentRule
		andProducesNode: aCollection! !

!S2JParserTest methodsFor: 'accessing' stamp: 'ag 11/21/2010 14:14'!
currentRule
	^ currentRule! !

!S2JParserTest methodsFor: 'accessing' stamp: 'ag 11/21/2010 14:14'!
currentRule: aSymbol
	currentRule := aSymbol! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 20:37'!
node: aString
	
	^ S2JNode named: aString with: {}! !

!S2JParserTest methodsFor: 'helpers' stamp: 'ag 11/21/2010 19:27'!
node: aString with: aCollection
	
	^ S2JNode named: aString with: aCollection! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 20:02'!
testBinarySelector
	"Would be cool to find all the binary messages in the image and test them..."
	
	self
		currentRule: #binarySelector;
		
		"special cases: single pipe and single minus"
		assertMatches: '|' andProduces: #|;
		assertMatches: '-' andProduces: #-;
		
		"names starting with a minus"
		assertMatches: '-*' andProduces: #-*;
		assertMatches: '-***' andProduces: #-***;
		assertMatches: '-*%&' andProduces: #-*%&;
		assertMatches: '-*/\' andProduces: #-*/\;
		assertMatches: '-!!?' andProduces: #-!!?;
		assertMatches: '-==' andProduces: #-==;
		
		"mixed characters, check each one"
		assertMatches: '*+' andProduces: #*+;
		assertMatches: '/=\' andProduces: #/=\;
		assertMatches: '~,,,<' andProduces: #~,,,<;
		assertMatches: '>@' andProduces: #>@;
		assertMatches: '%&' andProduces: #%&;
		assertMatches: '?!!' andProduces: #?!!;
		assertMatches: '``' andProduces: #``;
		assertMatches: '+??=' andProduces: #+??=;
		assertMatches: ',!!!!,' andProduces: #,!!!!,;
		
		"Not allowed things"
		assertFails: '';
		assertFails: ' ';
		assertFails: 'w';
		assertFails: '2';
		assertFails: '--';
		assertFails: '||';
		assertFails: '|--';
		assertFails: '%--';
		assertFails: '%||';
		assertFails: '&-&-';
		assertFails: '- %';
		assertFails: ' -';
		assertFails: '& &';
		assertFails: 'w a';
		assertFails: '- -';
		assertFails: '#a a';
		assertFails: '#aa 2'.! !

!S2JParserTest methodsFor: 'test-complex' stamp: 'ag 11/21/2010 20:59'!
testBlock
	
	self
		currentRule: #block;
		
		assertMatches: '[]' andProducesNode: { };
		assertMatches: '[  ]' andProducesNode: { };
		assertMatches: '[||]' andProducesNode: { self node: 'temporaries' };
		assertMatches: '[:a|||]' andProducesNode: { self node: 'arguments' with: { 'a' }. self node: 'temporaries' };
		assertMatches: '[||a]' andProducesNode: { self node: 'temporaries'. self node: 'statements' with: { self node: 'expression' with: { 'a' } } };
		assertMatches: '[:a|||a]' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'temporaries'. self node: 'statements' with: { self node: 'expression' with: { 'a' } } };
		assertMatches: '[:a||a|a]' andProducesNode: 
			{ self node: 'arguments' with: { 'a' }. self node: 'temporaries' with: { 'a' }. self node: 'statements' with: { self node: 'expression' with: { 'a' } } };
		assertMatches: '[:a|a|a|a]' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'statements' with:
				{ self node: 'expression' with: { self node: 'binaryExpression' with: { self node: 'binaryExpression' with: { 'a'. #|. 'a' }. #|. 'a' }}}};
		assertMatches: '[ :a ]  ' andProducesNode: { self node: 'arguments' with: { 'a' } };
		assertMatches: '   [ :a ]  ' andProducesNode: { self node: 'arguments' with: { 'a' } };
		assertMatches: '[ :a :b ]  ' andProducesNode: { self node: 'arguments' with: { 'a'. 'b' } };
		assertMatches: '[ :a     :b | ]' andProducesNode: { self node: 'arguments' with: { 'a'. 'b' }  };
		assertMatches: '  [   :a |   |  a  |   ]  ' andProducesNode: { self node: 'arguments' with: { 'a' }. self node: 'temporaries' with: { 'a' } };
		assertMatches: '[  :a  | a ]  ' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'statements' with: { self node: 'expression' with: { 'a' } } };
		assertMatches: '[ :a  | a. ^ 1 ]' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'statements' with:
				{ self node: 'expression' with: { 'a' }. self node: 'returnExpression' with: { 1 }}};
		assertMatches: '[  :a  |  1  ]' andProducesNode:
			{ self node: 'arguments' with: { 'a' }. self node: 'statements' with: { self node: 'expression' with: { 1 } } };
		assertMatches: '[  :b  |   []  ]' andProducesNode:
			{ self node: 'arguments' with: { 'b' }. self node: 'statements' with: { self node: 'expression' with: { self node: 'block' } } };
		
		assertFails: '[]]';
		assertFails: '[[]';
		assertFails: '[::a]';
		assertFails: '[ :a | b';
		assertFails: '[ :a |||| b]';
		assertFails: '[ | :a | b]';
		assertFails: '[ || :a | b]';
		assertFails: '[ :a || b]';
		assertFails: '[ :a | :b  ]';
		assertFails: '[ :a || :b|]';
		assertFails: '[ :a | b| :a]';
		assertFails: '[ :a :1 | b]'.! !

!S2JParserTest methodsFor: 'test-complex' stamp: 'ag 11/21/2010 21:46'!
testBraceExpression
	
	self
		currentRule: #braceExpression;
		
		assertMatches: '{}' andProducesNode: { };
		assertMatches: '  {  }   ' andProducesNode: { };
		assertMatches: '{1}' andProducesNode: { self node: 'expression' with: { 1 } };
		assertMatches: '  {  11 }  ' andProducesNode: { self node: 'expression' with: { 11 } };
		assertMatches: '  {  11. }  ' andProducesNode: { self node: 'expression' with: { 11 } };
		assertMatches: '{ 1.   2}' andProducesNode: { self node: 'expression' with: { 1 }. self node: 'expression' with: { 2 } };
		assertMatches: '{1.2}' andProducesNode: { self node: 'expression' with: { 1 }. self node: 'expression' with: { 2 } };
		assertMatches: '{1. ''.'' .}' andProducesNode: { self node: 'expression' with: { 1 }. self node: 'expression' with: { '.' } };
		assertMatches: '{1  .  2  .}' andProducesNode: { self node: 'expression' with: { 1 }. self node: 'expression' with: { 2 } };
		
		assertFails: '.{}.';
		assertFails: '{1 .. 2}';
		assertFails: '{ 1  . 2 ..}';
		assertFails: '{.1.2}';
		assertFails: '{..1.}'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 20:01'!
testCharacter
	
	self
		currentRule: #character;
		
		assertMatches: '$c' andProduces: $c;
		assertMatches: '$&' andProduces: $&;
		assertMatches: '$ ' andProduces: $ ;
		assertMatches: '$`' andProduces: $`;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: 'a';
		assertFails: '`';
		assertFails: ' $a';
		assertFails: '$$$';
		assertFails: '$  a';
		assertFails: '$av';
		assertFails: '$ a'.! !

!S2JParserTest methodsFor: 'test-complex' stamp: 'ag 11/21/2010 20:07'!
testExpression
	
	#todo.
	"TODO"
	
	self
		currentRule: #expression;
		
		assertMatches: 'a' andProducesNode: { 'a' };
		
		assertFails: 'a : = b'.! !

!S2JParserTest methodsFor: 'test-complex' stamp: 'ag 11/21/2010 20:03'!
testExpressionAssignment
	
	self
		currentRule: #expression;
		
		assertMatches: 'a' andProducesNode: { 'a' };
		assertMatches: 'a := b' andProducesNode: { self node: 'variableAssignments' with: { 'a' }. 'b' };
		assertMatches: '  a   :=   b  ' andProducesNode: { self node: 'variableAssignments' with: { 'a' }. 'b' };
		assertMatches: 'a := b := c := 1' andProducesNode: { self node: 'variableAssignments' with: { 'a'. 'b'. 'c' }. 1 };
		assertMatches: '  a    :=  b  :=  c :=   1' andProducesNode: { self node: 'variableAssignments' with: { 'a'. 'b'. 'c' }. 1 };
		assertMatches: 'a _ b' andProducesNode: { self node: 'variableAssignments' with: { 'a' }. 'b' };
		
		assertFails: 'a : = b';
		assertFails: ' := b';
		assertFails: 'a := ';
		assertFails: 'a := b :=';
		assertFails: 'a := b := 1 := a';
		assertFails: 'a := := b   ';
		assertFails: 'a _ _ b';
		assertFails: 'a __ b';
		assertFails: 'a :=:= b';
		assertFails: 'a := b v := a';
		assertFails: 'a  b := a '.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 15:36'!
testIdentifier
	
	self
		currentRule: #identifier;
		
		assertMatches: 'abc' andProduces: 'abc';
		assertMatches: 'a1' andProduces: 'a1';
		assertMatches: '1' andProduces: '1';
		assertMatches: 'aaaaaaaa1233333' andProduces: 'aaaaaaaa1233333';
		assertMatches: '' andProduces: '';
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '|';
		assertFails: ':';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: '123';
		assertFails: 'abc';
		assertFails: 'abc:';
		assertFails: 'abc<<'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 15:34'!
testKeywordSelector
	
	self
		currentRule: #keywordSelector;
		
		assertMatches: 'abc:' andProduces: #abc:;
		assertMatches: 'a1:' andProduces: #a1:;
		assertMatches: '1:' andProduces: #1:;
		assertMatches: 'aaaaaaaa1233333:' andProduces: #aaaaaaaa1233333:;
		assertMatches: ':' andProduces: #:;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: '123';
		assertFails: 'abc';
		assertFails: 'abc';
		assertFails: 'abc<<'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 15:56'!
testNumber
	
	"TODO extend test, when numberS2J gets extended (e.g. -2 0.2 etc.)"
	
	self
		currentRule: #numberS2J;
		
		assertMatches: '1' andProduces: 1;
		assertMatches: '0' andProduces: 0;
		assertMatches: '123' andProduces: 123;
		assertMatches: '112345678' andProduces: 112345678;
		assertMatches: '000' andProduces: 0;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: 'abc';
		assertFails: '<<'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 20:01'!
testString
	
	"TODO extend test, if stringS2J gets extended to support '' inside of a string"
	
	self
		currentRule: #stringS2J;
		
		assertMatches: '''a''' andProduces: 'a';
		assertMatches: '''''' andProduces: '';
		assertMatches: '''123''' andProduces: '123';
		assertMatches: '''    ''' andProduces: '    ';
		assertMatches: '''"$%&''' andProduces: '"$%&';
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '#abc';
		assertFails: ''' ''a';
		assertFails: ' ''''';
		assertFails: 'a''''';
		assertFails: '''as';
		assertFails: 'as''';
		assertFails: ''' '' ''';
		assertFails: '''a '' a''';
		assertFails: '''/'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 20:01'!
testSymbol
	
	"TODO extend test, if stringS2J gets extended to support '' inside of a string"
	
	self
		currentRule: #symbolS2J;
		
		assertMatches: '#abc' andProduces: #abc;
		assertMatches: '#abc1' andProduces: #abc1;
		assertMatches: '#a' andProduces: #a;
		assertMatches: '#''a''' andProduces: #'a';
		assertMatches: '#''''' andProduces: #'';
		assertMatches: '#''123''' andProduces: #'123';
		assertMatches: '#''    ''' andProduces: #'    ';
		assertMatches: '#''"$%&''' andProduces: #'"$%&';
		
		assertFails: '';
		assertFails: ' ';
		assertFails: 'abc';
		assertFails: '#""';
		assertFails: '#'' ''a';
		assertFails: '# ''''';
		assertFails: '#a''''';
		assertFails: '#''as';
		assertFails: '#as''';
		assertFails: '#'' '' ''';
		assertFails: '#''a '' a''';
		assertFails: '#123';
		assertFails: '###';
		assertFails: '#''/'.! !

!S2JParserTest methodsFor: 'test-tokens' stamp: 'ag 11/21/2010 16:06'!
testUnarySelector
	"Would be cool to find all the unary messages in the image and test them..."
	
	self
		currentRule: #unarySelector;
		
		assertMatches: 'abc' andProduces: #abc;
		assertMatches: 'a1' andProduces: #a1;
		assertMatches: '1' andProduces: #1;
		assertMatches: 'aaaaaaaa1233333' andProduces: #aaaaaaaa1233333;
		assertMatches: '' andProduces: #;
		
		assertFails: '';
		assertFails: ' ';
		assertFails: '|';
		assertFails: ':';
		assertFails: '#abc';
		assertFails: '1abc';
		assertFails: '123';
		assertFails: 'abc';
		assertFails: 'abc:';
		assertFails: 'abc<<'.! !

OMeta2 subclass: #S2JParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JParser commentStamp: 'ag 11/21/2010 19:26' prior: 0!
tokens
	Rules, that read one single element in squeal source-code. None of them returns an S2JNode, all are parsed to the
	according Squeak-object (e.g. Number, Symbol, String).
	Spaces before and after the tokens in the source-code are NOT read!!
	This is because in some cases, tokens must be read exactly, without emitting any spaces.
	The things returned from here must not be converted anymore, especially the <> operator is unnecessary.
	
complex elements
	Bigger parts of squeak source-code. Most return an S2JNode (using #node:with:), which contains further complex elements or object read
	by rules in the tokens-protocol. As convention, the nodes (except for helper-rules like primary and messagePattern) should be named like the rules.
	Each rule is supposed to read spaces before applying itself!! This way, rules from this protocol can be called serially, without
	reading spaces inbetween. However, spaces must be read between single tokens.
	
helpers
	Squeak-methods, that are called from within ometa-rules (or helper-rules).
	OMeta2 return-parts of the rules (righthand of ->) cannot contain multiple expressions (?)...
!

!S2JParser class methodsFor: 'matching' stamp: 'ag 11/21/2010 19:58'!
matchString: aString with: aRule
	
	| matcher result |
	matcher := self matcherOn: aString readStream.
	result := matcher
		apply: #empty;
		apply: aRule.
	matcher
		apply: #spaces;
		apply: #end.
	^ result! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:11'!
arrayLiteral =
	
	'thiswillfail' #todo! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 20:30'!
binaryExpression =
	
	(binaryExpression | primary | unaryExpression):receiver
	spaces binarySelector:selector
	(primary | unaryExpression):argument
	
	-> [ self node: 'binaryExpression' with: { receiver. selector. argument} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
binarySelector =
	
	<$| | ((specialCharacter | $-) (specialCharacter* | empty))>:selector
	-> [ selector asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 20:44'!
block =
	
	ch($[)
		(( (ch($:) identifier:argumentName -> [ argumentName ])+:arguments (ch($|):argumentSeparator | empty) ) | empty)
		(temporaries:temporaries | empty)
		(statements:statements | empty)
	ch($])
	
	// Forbidden (and not covered above): Arguments without separator bar, but with statements or temporaries
	?[ argumentSeparator notNil | arguments isNil | ((temporaries isNil or: [temporaries isEmpty]) & (statements isNil or: [statements isEmpty])) ]
	
	-> [ self node: 'block' withoutNils: {
			self node: 'arguments' withoutNilsNotEmpty: arguments.
			temporaries. statements } ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 21:38'!
braceExpression =
	
	ch(${)
		(expression:expr ch($.) -> [ expr ] )*:expressions
		(expression:lastExpression | empty)
	ch($})
	
	-> [ self braceExpressionWith: expressions and: lastExpression ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/21/2010 21:38'!
braceExpressionWith: expressions and: lastExpression
	
	^ self node: 'braceExpression' withoutNils: 
		((expressions
			ifNotNil: [ expressions ]
			ifNil: [ {} ])
				asOrderedCollection
					add: lastExpression;
					yourself)! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:11'!
cascadedExpression =
	
	'thiswillfail' #todo! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/21/2010 16:50'!
ch :c =
	// This reads a given element and all spaces before it and return the element
	// ch was supposed to mean character, but this can be applied to any object
	// Called like this: ch($=)
	
	spaces exactly(c)
	-> [ c ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 15:50'!
character =
	// The method in the super class is #char, don't confuse it!!
	
	$$ <anything>:value
	-> [ value asCharacter ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 18:37'!
expression =
	
	(((spaces identifier:identifier (token(':=') | ch($_))) -> [ identifier ])*:variableNames 
	| empty)
	(unaryExpression | binaryExpression | keywordExpression | primary | cascadedExpression):expressionBody
	
	// Create an extra node for the assignments, if any
	-> [ self node: 'expression' withoutNils: { 
			(self node: 'variableAssignments' withoutNilsNotEmpty: variableNames).
			expressionBody} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 20:04'!
identifier =
	
	<letter letterOrDigit*>! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:11'!
keywordExpression =
	
	'thiswillfail' #todo! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/21/2010 18:59'!
keywordMessagePattern: selectorAndArgumentNames
	"selectorAndArgumentNames contains associations, the part of the selector and the argument name.
	Resulting S2JNode will have the selector as symbol on the first position and the argumentNames at the other positions."
	
	| selector result |
	selector := ''.
	result := self node: 'keyword' with: {}.
	selectorAndArgumentNames do: [ :each |
		selector := selector, each key.
		result add: each value].
	^ result
		addFirst: selector asSymbol;
		yourself! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:15'!
keywordSelector =
	
	<identifier $:>:selector
	-> [ selector asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:01'!
messagePattern =
	
	spaces identifier:unarySelector
	-> [ self node: 'unary' with: { unarySelector } ]
|	spaces binarySelector:selector spaces identifier:argumentName
	-> [ self node: 'binary' with: { selector. argumentName } ]
|	((spaces keyword:selectorPart spaces identifier:argumentName) -> [ selectorPart -> argumentName ])*:selectorAndArgumentNames
	-> [ self keywordMessagePattern: selectorAndArgumentNames ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:02'!
method =
	
	messagePattern:head
	(temporaries:temps | empty)
	(primitiveDeclaration:primitive | empty)
	(statements:statements | empty)
	
	-> [ self node: 'method' withoutNils: { head. temps. primitive. statements } ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 23:33'!
node: aString with: anOrderedCollection
	
	^ S2JNode named: aString with: anOrderedCollection! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 23:33'!
node: aString withoutNils: anOrderedCollection
	
	^ S2JNode named: aString withoutNils: anOrderedCollection! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/18/2010 23:34'!
node: aString withoutNilsNotEmpty: anOrderedCollection
	
	^ S2JNode named: aString withoutNilsNotEmpty: anOrderedCollection! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 19:12'!
numberS2J =
	// Called that way because of super-class-method
	// TODO numbers can be expressed in several different ways!!
	(#todo | empty)
	
	<(anything:n ?[n isDigit])+>:result
	-> [ result asNumber ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 21:49'!
primary =
	// something, that evaluates to an object
	
	spaces (identifier | numberS2J | symbolS2J | character | stringS2J):result -> [ result ]
|	arrayLiteral
|	block
|	braceExpression
|	token('(') expression:expr token(')') -> [ expr ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:26'!
primitiveDeclaration =
	
	// TODO primitive can also take the form <primitive: methodNameString module: moduleNameString>
	(#todo | empty)
	
	ch($<)
		token('primitive:')
		spaces number:primitiveNumber ?[ primitiveNumber isInteger ]
	ch($>)
	-> [ self node: 'primitiveDeclaration' with: { primitiveNumber } ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:46'!
specialCharacter =
	
	$+ | $* | $/ | $\ | $~ | $< | $> | $= | $@ | $% | $& | $? | $!! | $` | $,! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:08'!
statements =
	
	(expression:expr ch($.) -> [ expr ] )*:expressions
	
	(((ch($^):returnSign | empty)
		expression:lastExpression (ch($.) | empty) ~expression)
	| empty)
	
	-> [ self statementsWithExpressions: expressions lastExpression: lastExpression returnSign: returnSign ]! !

!S2JParser methodsFor: 'helpers' stamp: 'ag 11/21/2010 20:35'!
statementsWithExpressions: expressions lastExpression: lastExpression returnSign: returnSign
	"Build the contents of a statements-node from the information gathered while parsing 'statements'"
	
	"Rename the last expression, if it was a return-expression"
	lastExpression notNil & returnSign notNil ifTrue: [ lastExpression name: 'returnExpression' ].
	expressions add: lastExpression.
	^ self node: 'statements' withoutNilsNotEmpty: expressions! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 20:04'!
stringS2J =
	// Called that way because of super-class-method
	
	$' <(~$' anything)*>:value $'
	-> [ value ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 15:53'!
symbolS2J =
	// Called that way because of super-class-method
	
	$# (identifier | stringS2J):value
	-> [ value asSymbol ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:10'!
temporaries =
	
	ch($|)
		(spaces identifier:identifier -> [ identifier ])*:variableNames
	ch($|)
	-> [ self node: 'temporaries' with: variableNames ]! !

!S2JParser methodsFor: 'complex elements' stamp: 'ag 11/21/2010 19:09'!
unaryExpression =
	
	(primary | unaryExpression):receiver identifier:message
	-> [ self node: 'unaryExpression' with: {receiver. message} ]! !

!S2JParser methodsFor: 'tokens' stamp: 'ag 11/21/2010 13:13'!
unarySelector =
	
	identifier:selector
	-> [ selector asSymbol ]! !

OMeta2 subclass: #S2JTranslator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JTranslator class methodsFor: 'as yet unclassified' stamp: 'ag 11/21/2010 21:08'!
squeakCharacters
	
	^ #( 
	$: 
	"Special characters in binary method names"
	$+ $* $/ $\ $~ $< $> $= $@ $% $& $? $!! $` $, 
	$| $-)! !

!S2JTranslator class methodsFor: 'as yet unclassified' stamp: 'ag 11/21/2010 21:32'!
squeakToJS: aStringOrSymbol
	"This is based on the fact, that the character $_ is forbidden in squeak, but allowed in JS."
	
	"TODO will not suffice -- converting a method-name backwards will be hard (______ can mean any number of combinations of these chars).
	Will need to do some prime-number-magic or find second character, that is forbidden in squeak, but allowed in js... ?"
	
	| result |
	result := ''.
	aStringOrSymbol asString do: [ :each | | index |
		index := self squeakCharacters indexOf: each.
		index ~= 0
			ifTrue: [ 1 to: index do: [ result := result, '_' ]]
			ifFalse: [ result := result + each asString ]].
	^ result! !

!S2JTranslator methodsFor: 'translate' stamp: 'ag 11/21/2010 21:04'!
methodCall
	
	! !

!S2JTranslator methodsFor: 'helpers' stamp: 'ag 11/21/2010 15:21'!
node :aString =
	
	// This does not create an S2JNode, but reads one from the input stream and returns it!!
	// For this purpose, this is an ometa-rule with one argument, call like this: node('binarySelector')
	// Other things, that can be found in the result of S2JParser are symbols, strings, numbers etc. (already implemented in superclass)
	// Collections, that are not S2JNodes, should not occur...!!
	
	anything:node ?[node class = S2JNode and: [node name = aString]]
	-> [ node ]! !

OrderedCollection subclass: #S2JNode
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Squeak2JS'!

!S2JNode commentStamp: 'ag 11/18/2010 13:09' prior: 0!
This is only a named collection, that is supposed to hold further collections or strings to form a tree!

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:08'!
named: aString with: anOrderedCollection
	
	^ self new
		name: aString;
		addAll: anOrderedCollection;
		yourself! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/18/2010 13:58'!
named: aString withoutNils: anOrderedCollection
	
	anOrderedCollection ifNil: [ ^ nil ].
	^ self
		named: aString
		with: (anOrderedCollection reject: [ :each | each isNil])! !

!S2JNode class methodsFor: 'instance creation' stamp: 'ag 11/21/2010 19:15'!
named: aString withoutNilsNotEmpty: anOrderedCollection
	"Returns nil, if resulting node would be empty"
	
	| result |
	result := self named: aString withoutNils: anOrderedCollection.
	^ (result isNil or: [result isEmpty])
		ifTrue: [ nil ]
		ifFalse: [ result ]! !

!S2JNode methodsFor: 'comparing' stamp: 'ag 11/21/2010 19:19'!
= anObject
	
	^ super = anObject and: 
		[ name = anObject name ]! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name
	
	^ name! !

!S2JNode methodsFor: 'accessors' stamp: 'ag 11/18/2010 13:07'!
name: aString
	
	name := aString! !

!S2JNode methodsFor: 'printing' stamp: 'ag 11/19/2010 00:39'!
printString
	
	| result first |
	first := true.
	result := name, ' ('.
	self do: [ :each |
		first ifTrue: [first := false] ifFalse: [result := result, ', '].
		result := result,  each printString].
	^ result, ')'! !
